/*
    This file is part of "now", a simple calendar tool
    Copyright (C) 2012-2015  Laurent Poirrier

    "now" is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include "utf8.h"
#include "terminal.h"


#define DEBUG(s...)	fprintf(stderr, s)


// *************************************************************************
// 
// *************************************************************************
static volatile unsigned long term_signaled_flag = 0;

static void term_winch(int sig)
{
	if (sig != SIGWINCH) {
		DEBUG("term_winch(): called with signal %d\n", sig);
	}
	
	term_signaled_flag++;
}


static int term_install_signals()
{
	struct sigaction siga;
	
	siga.sa_handler = term_winch;
	memset(&siga.sa_mask, 0, sizeof(siga.sa_mask));
	siga.sa_flags = 0;
	
	if (sigaction(SIGWINCH, &siga, NULL)) {
		DEBUG("sigaction(SIGWINCH): %s\n", strerror(errno));
		return(-1);
	}
	
	return(0);
}

// *************************************************************************
// 
// *************************************************************************
int term_open(term *t)
{
	struct termios tio;
	
	if (term_install_signals())
		return(-1);

	t->fd = open("/dev/tty", O_RDWR);
	t->signaled = 0;
	
	if (t->fd == -1) {
		DEBUG("/dev/tty: %s\n", strerror(errno));
		return(-1);
	}
	
	if (tcgetattr(t->fd, &tio)) {
		DEBUG("/dev/tty: tcgetattr(): %s\n", strerror(errno));
		close(t->fd);
		return(-1);
	}
		
	
	tio.c_lflag &= ~(ECHO | ICANON | ISIG);
	tio.c_cc[VMIN] = 1;
	tio.c_cc[VTIME] = 0;
	
	if (tcsetattr(t->fd, TCSAFLUSH, &tio)) {
		DEBUG("/dev/tty: tcsetattr(): %s\n", strerror(errno));
		close(t->fd);
		return(-1);
	}
	
	t->buff.n = 0;
	
	return(0);
}

void term_close(term *t)
{
	struct termios tio;

	if (tcgetattr(t->fd, &tio)) {
		DEBUG("/dev/tty: tcgetattr(): %s\n", strerror(errno));
	} else {
		tio.c_lflag |= (ECHO | ICANON | ISIG);
		tio.c_cc[VMIN] = 1;
		tio.c_cc[VTIME] = 0;
	
		if (tcsetattr(t->fd, TCSAFLUSH, &tio)) {
			DEBUG("/dev/tty: tcsetattr(): %s\n", strerror(errno));
		}
	}
	
	close(t->fd);
}

int term_getsize(term *t)
{
	struct winsize ws;
	
	if (ioctl(t->fd, TIOCGWINSZ, &ws) == -1) {
		DEBUG("/dev/tty: %s\n", strerror(errno));
		return(-1);
	}
	
	t->width = ws.ws_col;
	t->height = ws.ws_row;
	
	return(0);
}

int term_signaled(term *t)
{
	unsigned long s, r;
	
	s = term_signaled_flag;
	r = s - t->signaled;
	t->signaled = s;
	
	return(r);
}

// *************************************************************************
// 
// *************************************************************************
int term_write(term *t, void *buff, size_t sz)
{
	size_t o;
	ssize_t n;

	if (sz == 0)
		return(0);
		
	o = 0;
	while (o < sz) {
		n = write(t->fd, buff + o, sz - o);
		
		if (n < 0) {
			if (errno != EINTR) {
				DEBUG("term_write(): %s\n", strerror(errno));
				return(-1);
			}
		} else if (n == 0) {
			DEBUG("term_write(): hangup\n");
			return(-1);
		} else {
			o += n;
		}
	}
	
	return(0);
}

int term_print(term *t, char *buff)
{
	size_t sz = strlen(buff);
	return(term_write(t, buff, sz));
}


int term_goto(term *t, int x, int y)
{
	char buff[32];
	int n;
	
	n = snprintf(buff, sizeof(buff), "\033[%d;%dH", y + 1, x + 1);
	return(term_write(t, buff, n));
}

int term_font(term *t, int on, int off)
{
	char buff[64];
	int n;
	
	n = snprintf(buff, sizeof(buff), "%s%s%s%s%s%s",
		(on & TERM_DEFAULTS) ? "\033[0m" : "",
		(on & TERM_BOLD) ? "\033[1m" :
			((off & TERM_BOLD) ? "\033[21m" : ""),
		(on & TERM_DIM) ? "\033[2m" :
			((off & TERM_DIM) ? "\033[22m" : ""),
		(on & TERM_UNDERLINE) ? "\033[4m" :
			((off & TERM_UNDERLINE) ? "\033[24m" : ""),
		(on & TERM_BLINK) ? "\033[5m" :
			((off & TERM_BLINK) ? "\033[25m" : ""),
		(on & TERM_REVERSE) ? "\033[7m" :
			((off & TERM_REVERSE) ? "\033[27m" : "")
		);
	return(term_write(t, buff, n));
}

int term_color(term *t, int fg, int bg)
{
	int bm[8] = { 0, 4, 2, 6, 1, 5, 3, 7 };
	
	char buff[32];
	int n;
	
	n = snprintf(buff, sizeof(buff), "%s\033[%dm\033[%dm",
		(fg & 8) ? "\033[2m" : "\033[22m",
		30 + bm[fg & 7], 40 + bm[bg & 7]);
	return(term_write(t, buff, n));
}


// ****************************************************************************
// 
// ****************************************************************************
// We have to read both special console codes and UTF-8 characters, and
// it is a whole mess...
int term_getkey(term *t, term_keypress *k)
{
	size_t un;
	ssize_t rn;
	
	while (1) {
		if ((t->buff.n >= 1) && (t->buff.k[0] == '\033')) {
			un = (t->buff.n > 4) ? 4 : t->buff.n;
		} else {
			un = utf8_full(t->buff.k, t->buff.n);
		}
		
		if (un <= t->buff.n) {
			memcpy(k->k, t->buff.k, un);
			k->n = un;
			memmove(t->buff.k, t->buff.k + un, t->buff.n - un);
			t->buff.n -= un;
			return(0);
		}
	
		rn = read(t->fd, t->buff.k + t->buff.n,
			sizeof(t->buff.k) - t->buff.n);

		if (rn < 0) {
			if (errno != EINTR)
				DEBUG("term_getkey(): read(): %s\n",
					strerror(errno));
			k->n = 0;
			return(-1);
		}
	
		t->buff.n += rn;
	}
}

int term_is(term_keypress *k, char *s)
{
	size_t l;
	
	l = strlen(s);
	if (k->n != l)
		return(0);
	return(memcmp(k->k, s, k->n) == 0);
}

int term_is_char(term_keypress *k, char c)
{
	if (k->n != 1)
		return(0);
	
	return(k->k[0] == c);
}

int term_is_ctrl(term_keypress *k, char c)
{
	if ((c >= 'a') && (c <= 'z'))
		c = c + 'A' - 'a';
	
	return(term_is_char(k, c - '@'));
}

int term_is_key(term_keypress *k, int key)
{
	switch (key) {
	case K_ESC:
		return(term_is(k, "\033"));
	case K_UP:
		return(term_is(k, "\033[A"));
	case K_DOWN:
		return(term_is(k, "\033[B"));
	case K_LEFT:
		return(term_is(k, "\033[D"));
	case K_RIGHT:
		return(term_is(k, "\033[C"));
	case K_INS:
		return(term_is(k, "\033[2~"));
	case K_DEL:
		return(term_is(k, "\033[3~"));
	case K_PG_UP:
		return(term_is(k, "\033[5~"));
	case K_PG_DOWN:
		return(term_is(k, "\033[6~"));
	case K_HOME:
		return(term_is(k, "\033[H") || term_is(k, "\033OH")
			|| term_is(k, "\033[1~") || term_is(k, "\033[7~"));
	case K_END:
		return(term_is(k, "\033[F") || term_is(k, "\033OF")
			|| term_is(k, "\033[4~") || term_is(k, "\033[8~"));
	case K_BACK:
		return(term_is(k, "\177"));
	case K_ENTER:
		return(term_is(k, "\n"));
	case K_TAB:
		return(term_is(k, "\t"));
	default:
		return(0);
	}
}

