/*
    This file is part of "now", a simple calendar tool
    Copyright (C) 2012-2015  Laurent Poirrier

    "now" is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <string.h>
#include "memory.h"
#include "textedit.h"


// **************************************************************************
// 
// **************************************************************************
void ted_init(textedit *e, screen *scr, int flags)
{
	e->scr = scr;
	e->flags = flags;
	e->modified = 0;
	
	e->buff = tb_malloc(1);
	e->buff[0] = 0;

	e->x = e->y = e->width = e->height = 0;
	ted_cursor_reset(e);
}

// **********************************************************
void ted_clear(textedit *e)
{
	tb_free(e->buff);
}

// **********************************************************
void ted_cursor_reset(textedit *e)
{
	if (e->flags & TED_AT_END) {
		e->ox = e->oy = 0;
		e->cx = e->cy = 1 << 24;
	} else {
		e->ox = e->oy = e->cx = e->cy = 0;
	}
}

// **********************************************************
void ted_position(textedit *e, int x, int y, int w, int h)
{
	e->x = x;
	e->y = y;
	e->width = w;
	e->height = h;
}


// **************************************************************************
// 
// **************************************************************************
char *ted_get(textedit *e)
{
	return(e->buff);
}

void ted_set(textedit *e, char *s)
{
	size_t b;
	
	if (!s)
		s = "";
	
	b = strlen(s);
	e->buff = tb_realloc(e->buff, b + 1);
	memcpy(e->buff, s, b + 1);
	ted_cursor_reset(e);
	e->modified = 0;
}

void ted_set_modified(textedit *e, int m)
{
	e->modified = m;
}

int  ted_modified(textedit *e)
{
	return(e->modified);
}

// **************************************************************************
// 
// **************************************************************************
size_t ted_cursor_offset(textedit *e)
{
	int nx, ny, nx2;
	char *s;
	utf8_code c;
	size_t o, b, r;
	
	nx = ny = 0;
	s = e->buff;
	r = o = 0;
	
	if (e->cx < 0) {
		if (e->cy > 0) {
			e->cx = 1 << 24;
			e->cy--;
		} else {
			e->cx = 0;
		}
	}
	if (e->cy < 0)
		e->cy = 0;
	
	while (s[o]) {
		c = utf8_encode(s + o);
		b = utf8_bytes(s + o);
		
		if (c == '\n') {
			if (e->cy == ny)
				break;
			ny++;
			nx = 0;
		} else if (c == '\r') {
		} else if (c == '\t') {
			nx2 = (nx + 8) & (~7);
			
			if ((e->cy == ny) && (e->cx >= nx) && (e->cx < nx2)) {
				e->cx = nx;
				r = o;
			}
			nx = nx2;
		} else {
			if ((e->cy == ny) && (e->cx == nx))
				r = o;
			nx++;
		}
		
		o += b;
	}
	
	if (e->cy > ny) {
		e->cy = ny;
		r = o;
	}
	
	if ((e->cy == ny) && (e->cx >= nx)) {
		e->cx = nx;
		r = o;
	}
	
	// origin
	if (e->cx > e->ox + e->width - 1)
		e->ox = e->cx - e->width + 1;
	if (e->cy > e->oy + e->height - 1)
		e->oy = e->cy - e->height + 1;
	if (e->cx < e->ox)
		e->ox = e->cx;
	if (e->cy < e->oy)
		e->oy = e->cy;
	
	return(r);
}

void ted_insert(textedit *e, size_t offs, char *s, size_t b)
{
	size_t b0;
	
	b0 = tb_size(e->buff);
	e->buff = tb_realloc(e->buff, b0 + b);
	
	memmove(e->buff + offs + b, e->buff + offs, b0 - offs);
	memcpy(e->buff + offs, s, b);
}

void ted_delete(textedit *e, size_t offs)
{
	size_t b, b0;
	
	b0 = tb_size(e->buff);
	
	if (offs >= b0 - 1)
		return;
	
	b = utf8_bytes(e->buff + offs);
	memmove(e->buff + offs, e->buff + offs + b, b0 - (offs + b));
	e->buff = tb_realloc(e->buff, b0 - b);
}

// **************************************************************************
// 
// **************************************************************************
void ted_paint_internal(textedit *e, int fl)
{
	int nx, ny;
	char *s;
	int mem_n;
	utf8_code c;

	for (ny = 0; ny < e->height; ny++) {
		screen_multi(e->scr, e->x, e->y + ny, e->width,
			fl, SCREEN_BLANK);
	}
	
	nx = ny = 0;
	s = e->buff;
	mem_n = 0;
	
	while (*s) {
		if (mem_n > 0) {
			mem_n--;
		} else {
			c = utf8_encode(s);
			s += utf8_bytes(s);
		}
		
		if (c == utf8_encode("\n")) {
			ny++;
			nx = 0;
		} else if (c == utf8_encode("\r")) {
		} else if (c == utf8_encode("\t")) {
			c = ' ';
			mem_n = 8 - (nx & 7);
		} else {
			if ((nx - e->ox >= 0) && (nx - e->ox < e->width)
			&&  (ny - e->oy >= 0) && (ny - e->oy < e->height)) {
				screen_multi(e->scr,
					e->x + nx - e->ox, e->y + ny - e->oy,
					1, fl, c);
			}
			nx++;
		}
	}
	
	//screen_atf(e->scr, 0, 0, 40, 0, "(%d, %d): %zd",
	//	e->cx, e->cy, ted_cursor_offset(e));
}

// **************************************************************************
// 
// **************************************************************************
void ted_cursor_paint(textedit *e, term *tty, int active)
{
	if (!active)
		return;
	
	ted_cursor_offset(e);
	term_goto(tty, e->x + e->cx - e->ox, e->y + e->cy - e->oy);
}

// **************************************************************************
// 
// **************************************************************************
int ted_events(textedit *e, term_keypress *k)
{
	size_t o;
	
	if (term_is_key(k, K_ESC)) {
		if (e->flags & TED_MULTILINE)
			return(TED_CONFIRM);
		else
			return(TED_CANCEL);
	} else if (term_is_key(k, K_LEFT)) {
		e->cx--;
	} else if (term_is_key(k, K_RIGHT)) {
		o = ted_cursor_offset(e);
		if (e->buff[o] == '\t') {
			e->cx += 8;
		} else if (e->buff[o] == '\n') {
			e->cx = 0;
			e->cy++;
		} else {
			e->cx++;
		}
	} else if (term_is_key(k, K_UP)) {
		if (e->flags & TED_MULTILINE) {
			if ((e->cy == 0) && (e->flags & TED_UP_CYCLE))
				return(TED_PREV);
			else
				e->cy--;
		} else {
			return(TED_PREV);
		}
	} else if (term_is_key(k, K_DOWN)) {
		if (e->flags & TED_MULTILINE)
			e->cy++;
		else
			return(TED_NEXT);
	} else if (term_is_key(k, K_PG_UP)) {
		if (e->flags & TED_MULTILINE)
			e->cy -= 10;
		else if (e->flags & TED_MORELESS)
			return(TED_MORE);
		else
			return(TED_PREV);
	} else if (term_is_key(k, K_PG_DOWN)) {
		if (e->flags & TED_MULTILINE)
			e->cy += 10;
		else if (e->flags & TED_MORELESS)
			return(TED_LESS);
		else
			return(TED_NEXT);
	} else if (term_is_key(k, K_HOME)) {
		e->cx = 0;
	} else if (term_is_key(k, K_END)) {
		e->cx = 1 << 24;
	} else if (term_is_key(k, K_ENTER)) {
		if (e->flags & TED_MULTILINE) {
			o = ted_cursor_offset(e);
			ted_insert(e, o, "\n", 1);
			e->cy++;
			e->cx = 0;
		} else {
			return(TED_CONFIRM);
		}
	} else if (term_is_key(k, K_DEL)) {
		o = ted_cursor_offset(e);
		ted_delete(e, o);
		e->modified = 1;
	} else if (term_is_key(k, K_BACK)) {
		if ((e->cx > 0) || (e->cy > 0)) {
			e->cx--;
			o = ted_cursor_offset(e);
			ted_delete(e, o);
			e->modified = 1;
		}
	} else if (term_is_key(k, K_TAB)) {
		if ((e->flags & TED_MULTILINE)
		&&  (!(e->flags & TED_TAB_CYCLE))) {
			o = ted_cursor_offset(e);
			ted_insert(e, o, "\t", 1);
			e->cx = (e->cx + 8) & (~7);
			e->modified = 1;
		} else {
			return(TED_NEXT);
		}
	} else if ((k->k[0] < 0) || (k->k[0] >= ' ')) {
		o = ted_cursor_offset(e);
		ted_insert(e, o, k->k, k->n);
		e->cx++;
		e->modified = 1;
	}
	ted_cursor_offset(e);
	return(TED_CONTINUE);
}

