/*
    This file is part of "now", a simple calendar tool
    Copyright (C) 2012-2015  Laurent Poirrier

    "now" is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <string.h>
#include <errno.h>
#include "memory.h"
#include "bio.h"
#include "terminal.h"
#include "screen.h"
#include "theme.h"
#include "table.h"
#include "dateproc.h"
#include "event.h"
#include "ical.h"
#include "dialog.h"
#include "log.h"
#include "ui.h"

// ****************************************************************************
// 
// ****************************************************************************
static char *dayname[7] = { "Sunday", "Monday", "Tuesday", "Wednesday",
	"Thursday", "Friday", "Saturday" };

static char *monthname_short[13] = { "000", "Jan", "Feb", "Mar", "Apr",
	"May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

// ****************************************************************************
// 
// ****************************************************************************
static void paint_day(screen *s, cell *c, time_t day, event_list *l,
	int *sl, int *soffs, int *slast, event_handle **sev)
{
	time_t today;
	event_list dl;
	event_handle *h;
	dt_time r;
	char *mark;
	char c0, c1;
	int fl;
	int i;
	
	today = dt_daymid(dt_now());

	dt_localtime(day, &r);
	
	if ((sl) && (*sl < 0)) {
		fl = THEME_UI_DAYLINE_SEL;
	} else {
		fl = THEME_UI_DAYLINE;
	}
	
	if (sl) {
		if (day == today) {
			c0 = '*';
			c1 = '*';
		} else {
			c0 = '[';
			c1 = ']';
		}
	} else {
		if (day == today) {
			c0 = '=';
			c1 = '=';
		} else {
			c0 = ' ';
			c1 = ' ';
		}
	}
	
	screen_atf(s, c->x, c->y, c->width - 1, fl,
		"%c%s %2d%c",
		c0, monthname_short[r.mon], r.mday, c1);
	
	evlist_init(&dl);
	evlist_select(&dl, l, dt_daystart(day), dt_dayend(day));
	evlist_slowsort(&dl);
	h = dl.head;
	
	if ((sl) && (soffs) && (slast) && (sev) && (*sl >= 0)) {
		if (*sl > dl.size)
			*sl = dl.size;
		if (*sl < 0)
			*sl = 0;
		
		if (*soffs + c->height - 2 < *sl)
			*soffs = *sl - (c->height - 2);
		if (*soffs > *sl)
			*soffs = *sl;
		
		i = *soffs;
		while ((h) && (i > 0)) {
			h = h->next;
			i--;
		}
		
		*slast = dl.size;
		*sev = NULL;
	} else {
		soffs = NULL;
		slast = NULL;
		sev = NULL;
	}
	
	for (i = 0; i < c->height - 1; i++) {
		fl = THEME_UI_EVENT;
		
		if ((sl) && (*sl >= 0) && (*soffs + i == *sl)) {
			fl = THEME_UI_EVENT_SEL;
			if (h)
				*sev = h->orig;
		}
		
		if (i == c->height - 2) {
			fl |= THEME_UI_LAST_EVENT_MOD;
		}
		
		if (h) {
			if ((i == 0) && (soffs) && (*soffs > 0))
				mark = "^";
			else if ((i == c->height - 2) && (h->next))
				mark = "+";
			else
				mark = " ";
		
			if (h->ev->start < dt_daystart(day)) {
				screen_atf(s, c->x, c->y + 1 + i, c->width - 1,
				fl, "  ...%s%s",
				mark, h->ev->data[EVENT_SUMMARY]);
			} else {
				dt_localtime(h->ev->start, &r);
			
				screen_atf(s, c->x, c->y + 1 + i, c->width - 1,
					fl, "%02d:%02d%s%s", r.hour, r.min,
					mark, h->ev->data[EVENT_SUMMARY]);
			}
			h = h->next;
		} else {
			screen_at(s, c->x, c->y + 1 + i, c->width - 1, fl, "");
		}
	}
	
	evlist_clear(&dl);
}

// ****************************************************************************
// 
// ****************************************************************************
void ui_ev_edit(term *tty, screen *scr, screen *bkg,
	log *act, event_list *l, event_handle *selev, time_t day)
{
	event *ev;
	int r;
	
	if (selev) {
		ev = selev->ev;
	} else {
		ev = event_new();
		ev->start = day;
		ev->stop = ev->start + 3600;
	}
	
	r = ui_event(tty, scr, bkg, ev);
	
	if (selev) {
		if (r == 0) {
			log_appendf(act, "Mod: %4d-%02d-%02d %02d:%02d %s",
				dt_year(ev->start), dt_mon(ev->start),
				dt_mday(ev->start),
				dt_hour(ev->start), dt_min(ev->start),
				ev->data[EVENT_SUMMARY]);
		}
	} else {
		if (r == 0) {
			evlist_attach(l, ev);
			log_appendf(act, "New: %4d-%02d-%02d %02d:%02d %s",
				dt_year(ev->start), dt_mon(ev->start),
				dt_mday(ev->start),
				dt_hour(ev->start), dt_min(ev->start),
				ev->data[EVENT_SUMMARY]);
		} else {
			event_destroy(ev);
		}
	}
}

// ****************************************************************************
// 
// ****************************************************************************
void ui_ev_delete(term *tty, screen *scr, screen *bkg,
	log *act, event_list *l, event_handle *selev)
{
	int r;
	
	if (!selev)
		return;
	
	r = ui_selectn(tty, scr, bkg,
		selev->ev->data[EVENT_SUMMARY], 2,
		"Delete event.",
		"Keep event.");
	
	if ((r == -1) || (r == 1))
		return;
	
	log_appendf(act, "Del: %4d-%02d-%02d %02d:%02d %s",
		dt_year(selev->ev->start), dt_mon(selev->ev->start),
		dt_mday(selev->ev->start),
		dt_hour(selev->ev->start), dt_min(selev->ev->start),
		selev->ev->data[EVENT_SUMMARY]);

	event_destroy(selev->ev);
	evlist_detach(l, selev);
}

// ****************************************************************************
// 
// ****************************************************************************
int ui_ev_save(term *tty, screen *scr, screen *bkg,
	log *act, event_list *l, char *cal_path)
{
	char *xstr;
	
	if (b_create_dir_path(cal_path)) {
		xstr = xprintf("Could not create directory for %s: %s",
			cal_path, strerror(errno));
		ui_selectn(tty, scr, bkg, xstr, 1, "Continue without saving.");
		xfree(xstr);
		return(-1);
	}
	
	if (evlist_ical_save(l, cal_path)) {
		xstr = xprintf("Could not save calendar (%s: %s)",
			cal_path, strerror(errno));
		ui_selectn(tty, scr, bkg, xstr, 1, "Continue without saving.");
		xfree(xstr);
		return(-1);
	}
	
	log_crop(act, 5);
	return(0);
}

// ****************************************************************************
// 
// ****************************************************************************
int ui_ev_exit(term *tty, screen *scr, screen *bkg,
	log *act, event_list *l, char *cal_path)
{
	int r;
	
	if (act->n > 5) {
		log_modif(act, 0, "Save modifications and exit.");
		log_modif(act, 1, "Save modifications and continue editing.");
		log_modif(act, 2, "Discard modifications and exit.");
		
		while (1) {
			r = ui_select(tty, scr, bkg,
				"Unsaved modifications to the calendar:",
				act->n, act->lines);
		
			if ((r == 0) || (r == 1)) {
				// save
				if (ui_ev_save(tty, scr, bkg, act, l, cal_path))
					r = -1;
			}
			
			if ((r < 0) || (r == 1)) {
				// continue editing
				return(-1);
			}
			
			if ((r == 0) || (r == 2)) {
				// exit
				return(0);
			}
		}
	} else {
		r = ui_selectn(tty, scr, bkg,
			"Exit (no modifications to calendar):", 2,
			"Yes, exit.",
			"No, continue editing.");
		
		if (r == 1)
			return(-1);
		else
			return(0);
	}
}

// ****************************************************************************
// 
// ****************************************************************************
int ui_run(char *cal_path)
{
	// config
	int first_wday = 1;

	// output related stuff
	term tty;
	screen cur, bkg;
	table tab;
	int level, x, y, sx, sy, sl, soffs, slast;
	
	// input related stuff
	term_keypress k;
	
	// event related stuff
	event_list l;
	log act;

	// working variables
	time_t start, day, today, selday;
	event_handle *selev;
	int r;
	char *xstr;

	// Init log
	log_init(&act);
	log_appendf(&act, "");
	log_appendf(&act, "");
	log_appendf(&act, "");
	log_appendf(&act, "");
	log_appendf(&act, "----------------------------------------");

	// Init terminal input/output
	if (term_open(&tty))
		return(1);

	screen_init(&cur);
	screen_init(&bkg);
	table_init(&tab, 7, 6);
	level = 0;
	
	// Load
	evlist_init(&l);
	if (evlist_ical_load(&l, cal_path)) {
		xstr = xprintf("Could not read calendar (%s: %s)",
			cal_path, strerror(errno));
		r = ui_selectn(&tty, &cur, &bkg, xstr, 2,
			"Start with empty calendar, create file when saving.",
			"Exit now.");
		xfree(xstr);
		
		if (r != 0)
			level = -1;
	}

	// Init main loop
	today = dt_daymid(dt_now());
	start = dt_weekstart(dt_monthstart(today), first_wday);
	sy = (today + (3600 * 12) - start) / (3600 * 24);
	sx = sy % tab.width;
	sy = sy / tab.width;
	sl = -1;
	soffs = slast = 0;

	// Main loop
	while (level >= 0) {
		if ((level == 0) || (term_signaled(&tty))) {
			if (term_getsize(&tty))
				return(1);

			screen_resize(&cur, tty.width, tty.height);
			screen_resize(&bkg, tty.width, tty.height);
			
			screen_set(&cur, THEME_UI_BKG, SCREEN_BLANK);
			
			table_fit(&tab, 0, 1, tty.width, tty.height - 2);
		}
		
		level = 1;

		for (x = 0; x < tab.width; x++) {
			screen_atf(&cur,
				tab.elem[0][x].x, 0,
				tab.elem[0][x].width - 1,
				THEME_UI_TOPLINE,
				"%s", dayname[(x + first_wday) % 7]);
		}
		
		day = start;
		selday = 0;
		selev = NULL;
		for (y = 0; y < tab.height; y++) {
			for (x = 0; x < tab.width; x++) {
				if ((x == sx) && (y == sy)) {
					selday = day;
					paint_day(&cur, &tab.elem[y][x], day,
						&l, &sl, &soffs, &slast,
						&selev);
				} else {
					paint_day(&cur, &tab.elem[y][x], day,
						&l, NULL, NULL, NULL, NULL);
				}
				
				day = dt_nextday(day, 1);
			}
		}
		
		screen_atf(&cur, 0, tty.height - 1, tty.width - 1,
			THEME_UI_BOTTOMLINE, "%4d-%02d-%02d ",
			dt_year(selday), dt_mon(selday), dt_mday(selday));
		
		if (selev) {
			if (selev->ev->start < selday) {
				screen_atf(&cur, 11, tty.height - 1,
					tty.width - 12,
					THEME_UI_BOTTOMLINE,
					"| %s (%4d.%02d.%02d %02d:%02d - "
					"%4d.%02d.%02d %02d:%02d)",
					selev->ev->data[EVENT_SUMMARY],
					dt_year(selev->ev->start),
					dt_mon(selev->ev->start),
					dt_mday(selev->ev->start),
					dt_hour(selev->ev->start),
					dt_min(selev->ev->start),
					dt_year(selev->ev->stop),
					dt_mon(selev->ev->stop),
					dt_mday(selev->ev->stop),
					dt_hour(selev->ev->stop),
					dt_min(selev->ev->stop));
			} else {
				screen_atf(&cur, 11, tty.height - 1,
					tty.width - 12,
					THEME_UI_BOTTOMLINE,
					"| %02d:%02d %s (%d h %d min)",
					dt_hour(selev->ev->start),
					dt_min(selev->ev->start),
					selev->ev->data[EVENT_SUMMARY],
					(selev->ev->stop - selev->ev->start)
						/ 3600,
					((selev->ev->stop - selev->ev->start)
						% 3600) / 60);
			}
		}
		
		screen_tup(&cur, &tty, &bkg);
		term_goto(&tty, tty.width - 1, tty.height - 1);
		
		term_getkey(&tty, &k);

		if (sl < 0) {
			if (term_is_key(&k, K_ENTER)) {
				sl = 0;
			} else if (term_is_key(&k, K_UP)) {
				sy--;
			} else if (term_is_key(&k, K_DOWN)) {
				sy++;
			} else if (term_is_key(&k, K_LEFT)) {
				sx--;
			} else if (term_is_key(&k, K_RIGHT)) {
				sx++;
			} else if ((term_is_key(&k, K_ESC))
			|| (term_is_ctrl(&k, 'C'))) {
				if (!ui_ev_exit(&tty, &cur, &bkg, &act, &l,
				cal_path))
					level = -1;
				else
					level = 0;
			} else if (term_is_key(&k, K_PG_UP)) {
				start = dt_nextweek(start, -1);
			} else if (term_is_key(&k, K_PG_DOWN)) {
				start = dt_nextweek(start, 1);
			} else if (term_is_key(&k, K_HOME)) {
				sx = 0;
			} else if (term_is_key(&k, K_END)) {
				sx = tab.width - 1;
			}
		} else {
			if ((term_is_key(&k, K_ESC))
			||  (term_is_ctrl(&k, 'C'))) {
				sl = -1;
			} else if (term_is_key(&k, K_UP)) {
				sl--;
				if (sl < 0) {
					sy--;
					sl = 1 << 24;
				}
			} else if (term_is_key(&k, K_DOWN)) {
				sl++;
				if (sl > slast) {
					sy++;
					sl = 0;
				}
			} else if (term_is_key(&k, K_LEFT)) {
				sl = 0;
				sx--;
			} else if (term_is_key(&k, K_RIGHT)) {
				sl = 0;
				sx++;
			} else if (term_is_key(&k, K_HOME)) {
				sl = 0;
			} else if (term_is_key(&k, K_END)) {
				sl = 1 << 24;
			} else if (term_is_key(&k, K_PG_UP)) {
				start = dt_nextweek(start, -1);
			} else if (term_is_key(&k, K_PG_DOWN)) {
				start = dt_nextweek(start, 1);
			} else if (term_is_key(&k, K_ENTER)) {
				ui_ev_edit(&tty, &cur, &bkg,
					&act, &l, selev, selday);
				level = 0;
			} else if ((selev)
			&& ((term_is_key(&k, K_BACK)) || (term_is_key(&k, K_DEL)))) {
				ui_ev_delete(&tty, &cur, &bkg,
					&act, &l, selev);
				level = 0;
			}
		}
		if (sx < 0) {
			sx = tab.width - 1;
			sy--;
		}
		if (sx > tab.width - 1) {
			sx = 0;
			sy++;
		}
		
		if (sy < 0) {
			sy = 0;
			start = dt_nextweek(start, -1);
		}
		
		if (sy > tab.height - 1) {
			sy = tab.height - 1;
			start = dt_nextweek(start, 1);
		}
	}
	
	term_goto(&tty, tty.width - 1, tty.height - 1);
	term_font(&tty, TERM_DEFAULTS, 0);
	term_write(&tty, "\n", 1);

	screen_clear(&bkg);
	screen_clear(&cur);
	table_clear(&tab);
	term_close(&tty);
	
	evlist_destroy_all(&l);
	evlist_clear(&l);
	log_clear(&act);

	return(0);
}

